#!/usr/bin/env python3
# encoding=utf8
"""
Toolbox for ennio.

This file/module is for generic AWS helpers.
"""
import logging

from botocore.exceptions import ClientError
import boto3


def clean_log_groups(stack_id):
    """
    Ensure that all log groups in a removed stack are removed.

    Sounds silly but we've seen it happen before. For example, when we have
    defined a log group that goes with a lambda function, cfn will delete
    the log group first(because log group depends on the lambda), and at
    that time, if we have some requests goes to lambda, we will see a new
    log group created by AWS which will stay there and bug us when we want
    to deploy the stack again.
    """
    cfn_cli = boto3.client("cloudformation")
    log_cli = boto3.client("logs")

    paginator = cfn_cli.get_paginator("list_stack_resources")
    for response in paginator.paginate(StackName=stack_id):
        for resource in response["StackResourceSummaries"]:
            if resource["ResourceType"] != "AWS::Logs::LogGroup":
                continue
            log_group = resource["PhysicalResourceId"]
            try:
                log_cli.delete_log_group(logGroupName=log_group)
                # If the line above does not trigger an exception, we have
                # actually removed a log group. Log it down for reference.
                logging.info(f"Removed log group: {log_group}")
            except ClientError as err:
                code = err.response["Error"]["Code"]
                if code == "ResourceNotFoundException":
                    # log group not found, this is actually the expected
                    # behaviour.
                    continue
                raise


def empty_s3_bucket(bucket_name):
    """Remove all files in an S3 bucket."""
    s3cli = boto3.client("s3")

    # Determine whether this bucket exists.
    try:
        s3cli.head_bucket(Bucket=bucket_name)
    except ClientError as err:
        if err.response["Error"]["Code"] == "404":
            # The bucket does not exist, no need to empty it.
            logging.warning(f"Bucket does not exist: {bucket_name}.")
            return
        raise

    logging.warning(f"Emptying bucket: {bucket_name}.")
    bucket = boto3.resource("s3").Bucket(bucket_name)
    versioning = s3cli.get_bucket_versioning(Bucket=bucket_name).get("Status")
    if versioning in ["Enabled", "Suspended"]:
        bucket.object_versions.delete()
    else:
        bucket.objects.all().delete()



#!/usr/bin/env python3
# encoding=utf8
"""
Model definition for ennio.

Predefined commands:

compile: generate templates.
deploy: deploy specified stack
delete: delete specified stack
"""
from pathlib import Path
import argparse
import importlib
import inspect
import logging
import os
import sys

from botocore.exceptions import ClientError
import yaml

from .utils import InvalidConfigError, LazyBoto3Client


display_name = lambda name: name.replace("_", "-")
method_name = lambda name: name.replace("-", "_")


class EnnioConfig:
    """Represents a configuration yaml file for ennio."""

    def __init__(self, conf_file):
        conf_file = Path(conf_file)
        logging.debug(f"Reading conf file: {conf_file.resolve()}.")
        with open(conf_file) as fobj:
            self.data = yaml.load(fobj, Loader=yaml.SafeLoader)

        self.stacks = [stack["name"] for stack in self.data["stacks"]]
        self.prefixes = self.stacks + ["application"]
        self.validate()
        self.set_defaults()

    def __getitem__(self, name):
        """Pass the dict lookup to self.data."""
        return self.data[name]

    def validate(self):
        """Validate our DSL."""
        mandatory = ["application", "stacks", "deploy-steps"]
        if any(field not in self.data for field in mandatory):
            raise InvalidConfigError("Mandatory field missing in config.")

        if "name" not in self.data["application"]:
            raise InvalidConfigError(f"Undefined application name.")

        for step in self.data["deploy-steps"]:
            if not self.validate_step(step):
                raise InvalidConfigError(f"Invalid step defined: {step}.")

        extra_commands = self.data.get("extra-commands", [])
        for command in extra_commands:
            if not self.is_valid_method(command):
                raise InvalidConfigError(f"Invalid command defined: {command}.")

    def is_valid_method(self, method):
        """Return whether a method found in config is valid."""
        if "." not in method:
            return False

        if method.count(".") != 1:
            return False

        stack, _ = method.split(".")
        return stack in self.prefixes

    def validate_step(self, step):
        """Validate a single step in `deploy-steps` list."""
        if "stack" not in step and "operation" not in step:
            # miss both
            return False

        if "stack" in step and "operation" in step:
            # have both
            return False

        if "operation" in step and "." not in step["operation"]:
            # invalid operation specification.
            return False

        if "stack" in step:
            # return green if stack name found.
            return step["stack"] in self.stacks
        else:
            return self.is_valid_method(step["operation"])

    def set_defaults(self):
        """Setup default values for config."""
        for step in self.data["deploy-steps"]:
            step.setdefault("ignore_error", False)


class EnnioApplication:
    """Represents an application that has several cfn stacks as components."""

    NO_VERSION = "-1"

    cfn = LazyBoto3Client("cloudformation")
    ssm = LazyBoto3Client("ssm")

    def __init__(self, conf_file):
        self.config = EnnioConfig(conf_file)
        self.name = self.config["application"]["name"]
        self.namespace = os.environ.get("NAMESPACE", self.name)
        self.bucket = self.config["application"]["bucket"]
        self.yaml_tags = self.config["application"]["tags"]

        self.stacks = {}
        for config in self.config["stacks"]:
            mod_str, klass = config["class"].rsplit(".", 1)
            if mod_str not in sys.modules:
                mod = importlib.import_module(mod_str)
            stack_class = getattr(mod, klass)
            self.stacks[config["name"]] = stack_class(self, config)

        self.steps = self.parse_steps()
        self.extra_commands = {
            cmd.split(".")[1]: self.get_method(cmd)
            for cmd in self.config["extra-commands"]
        }

        self._version = None

    def parse_steps(self):
        """Parse the `deploy-steps` section in the config."""
        steps = []
        for config in self.config["deploy-steps"]:
            step = {}
            if "stack" in config:
                step = self.parse_stack_step(config)
            else:
                step = self.parse_operation_step(config)
            steps.append(step)
        return steps

    def get_method(self, method):
        """Get the method from method specification."""
        stack_name, name = method.split(".")
        if stack_name == "application":
            instance = self
        else:
            instance = self.stacks[stack_name]
        if not hasattr(instance, method_name(name)):
            raise InvalidConfigError(f"Invalid method `{method}` in config.")
        return getattr(instance, method_name(name))

    def parse_stack_step(self, config):
        """Parse step when the step involves a cloudformation stack."""
        no_op = lambda *args, **kwargs: logging.info("No operations performed.")
        stack = self.stacks[config["stack"]]

        step = {
            "name": config["stack"],
            "deploy": stack.deploy,
            "rollback": stack.rollback,
            "delete": stack.delete,
            "ignore_error": config["ignore_error"],
        }

        if os.environ.get("ENNIO_DELETE_ALL", "false").lower() == "true":
            step["delete"] = stack.delete
        elif "on_delete" in config:
            on_delete = config["on_delete"]
            if on_delete == "pass":
                step["delete"] = no_op
            else:
                step["delete"] = self.get_method(on_delete)
        return step

    def parse_operation_step(self, config):
        """Parse step when the step is an operation."""
        no_op = lambda *args, **kwargs: logging.info("No operations performed.")
        method = self.get_method(config["operation"])

        step = {
            "name": config["operation"],
            "deploy": method,
            "rollback": method,
            "delete": no_op,
            "ignore_error": config["ignore_error"],
        }
        if "on_delete" in config:
            on_delete = config["on_delete"]
            if on_delete == "pass":
                step["delete"] = no_op
            else:
                step["delete"] = self.get_method(on_delete)
        return step

    ##############################################
    # commandline entry point
    ##############################################
    def main(self):
        """
        Entry point.

        Parse command line arguments, find the method either from application
        class or a stack class, pass key word arguments to it and make it happen
        """
        parsed = self.parse_args()

        method = self.sub_commands[parsed.command]

        if not callable(method) and isinstance(method, str):
            print(method)
            return

        params = inspect.signature(method).parameters
        for key, value in params.items():
            if key not in parsed and value is not value.empty:
                raise argparse.ArgumentTypeError(
                    f"`{parsed.command}` need argument `--{display_name(key)}`."
                )
        kwargs = {
            name: getattr(parsed, name)
            for name in dir(parsed)
            if ((not name.startswith("_")) and name != "command")
        }
        method(**kwargs)

    def parse_args(self):
        """
        Argparser that can dynamically require arguments.

        We need to parse commandline args for different commands, these commands
        might be requiring different arguments. So we parse the command first,
        get the signature of the method, and add the parameters dynamically.
        """
        parser = argparse.ArgumentParser(
            usage=f"usage: {sys.argv[0]} [-h] command [options]"
        )
        actions = ", ".join(sorted(list(self.sub_commands.keys())))
        parser.add_argument(
            "command",
            help=f"Action to be carried out, Valid actions are: {actions}",
            metavar="command",
            choices=self.sub_commands,
        )
        parsed, unknown = parser.parse_known_args()
        for arg in unknown:
            if arg.startswith("--"):
                parser.add_argument(arg)
        return parser.parse_args()

    @property
    def sub_commands(self):
        """
        Only show a subset of all available commands.

        Each of the method defined in EnnioApplication is exposed to the
        commandline as sub commands. More over, parse_args would allow
        properties that return string to be called. However, it would be
        confusing to show all these properties and subcommand in the help
        message, so we need to filter here.
        """
        commands = {
            "delete-all": self.delete_all,
            "deploy-all": self.deploy_all,
        }
        for stack_name, stack in self.stacks.items():
            commands[f"deploy-{stack_name}"] = stack.deploy
            commands[f"delete-{stack_name}"] = stack.delete

        commands.update(self.extra_commands)
        return commands

    ##############################################
    # application version
    ##############################################
    @property
    def version_parameter(self):
        """
        Subclass should define a property named `namespace` for this to work.
        """
        return f"/application-versions/{self.namespace}"

    @property
    def version(self):
        """
        Get the version of the meta stack, return -1 if the version is not set.

        Meta stack means the stack(s) that are used by the application.
        For example, a typical ecs application has about 4 or more
        Cloudformation stacks, all these stacks is part of the meta stack.

        Subclass should define a property named `namespace` for this to work.
        """
        if self._version is None:
            logging.info(f"Getting version from SSM: {self.version_parameter}.")
            try:
                self._version = self.ssm.get_parameter(
                    Name=self.version_parameter
                )["Parameter"]["Value"]
            except ClientError as error:
                if error.response["Error"]["Code"] == "ParameterNotFound":
                    self._version = self.NO_VERSION
                else:
                    raise error
        logging.info(f"Got stack version: {self._version}.")
        return self._version

    @version.setter
    def version(self, _version):
        """
        Set the version of the meta stack.

        Subclass should define a property named `namespace` for this to work.
        """
        logging.info(f"Setting stack version: {_version}.")
        self.ssm.put_parameter(
            Name=self.version_parameter,
            Value=_version,
            Type="String",
            Overwrite=True,
        )
        self._version = _version

    @version.deleter
    def version(self):
        """
        Remove the version of the meta stack.

        Subclass should define a property named `namespace` for this to work.
        """
        logging.info(f"Removing stack version.")
        try:
            self.ssm.delete_parameter(Name=self.version_parameter)
            logging.info("Version removed from parameter store.")
        except ClientError as error:
            if not error.response["Error"]["Code"] == "ParameterNotFound":
                raise
            logging.info("Version does not exist in parameter store.")
        self._version = None

    ##############################################
    # Properties that can be used.
    ##############################################

    @property
    def tags(self):
        tags_ = self.yaml_tags
        return [{"Key": tag, "Value": tags_[tag]} for tag in tags_]

    ##############################################
    # Helper methods
    ##############################################
    def rollback_all(self, changed):
        """Rollback all changed stacks."""
        if self.version == self.NO_VERSION:
            logging.warning(f"Deploying for the first time, no rollback.")
            return

        steps = [step["name"] for step in changed]
        logging.info(f"Rolling back changes: {steps}.")
        for step in reversed(changed):
            logging.debug(f"running step: {step}")
            name = step["name"]
            logging.info(f"{name} rollback step started.")
            step["rollback"](self.version)
            logging.info(f"{name} rollback step finished.")
        logging.info(f"Rollback completed successfully.")

    ##############################################
    # Operations
    ##############################################
    def deploy_all(self, build):
        """Update all stacks in a transaction."""
        logging.info(f"Deploying {build}, current version: {self.version}.")

        changed = []
        for step in self.steps:
            logging.debug(f"running step: {step}")
            name = step["name"]
            try:
                logging.info(f"{name} deploy step started.")
                step["deploy"](build)
                logging.info(f"{name} deploy step finished.")
                changed.append(step)
            except Exception as err:
                logging.warning(f"{name} deploy step failed with: {err}")
                if step["ignore_error"]:
                    # Not going to add this step to changed, because we failed
                    # to change it.
                    logging.warning(f"Ignoring error for {name}. Error: {err}")
                    continue
                break
        else:
            # No break from for loop, all good.
            self.version = build
            logging.info(
                f"Deployment of application {self.name} completed successfully."
            )
            return

        if os.environ.get("ENNIO_NO_ROLLBACK", "false").lower() == "true":
            logging.warning(f"Rollback canceled by env var.")
        else:
            logging.warning(f"Reverting to build {self.version}.")
            self.rollback_all(changed)
        sys.exit(1)

    def delete_all(self):
        """Delete all stacks one by one."""
        logging.info(f"Removing stacks, current version: {self.version}.")

        for step in reversed(self.steps):
            logging.debug(f"running step: {step}")
            name = step["name"]
            try:
                logging.info(f"{name} delete step started.")
                step["delete"]()
                logging.info(f"{name} delete step finished.")
            except Exception as err:
                logging.warning(f"{name} delete step failed with: {err}")
                break

    def compile_all(self):
        """Compile all cfn templates and validate them all."""



#!/usr/bin/env python3
# encoding=utf8
"""Ennio is a framework for creating re-usable deployment scripts."""
import sys

from .app import EnnioApplication
from .stack import EnnioStack
from .utils import (
    LazyBoto3Client,
    require_aws,
    setup_logging,
    sleep,
)



#!/usr/bin/env python3
# encoding=utf8
"""Stack definition for ennio."""
from datetime import datetime
import functools
import logging
import os

from botocore.exceptions import ClientError

from .utils import format_changes, sleep, EmptyChangeSetError, LazyBoto3Client


class EnnioStack:
    """Represents a cloudformation stack."""

    cfn = LazyBoto3Client("cloudformation")
    ssm = LazyBoto3Client("ssm")
    log = LazyBoto3Client("logs")

    def __init__(self, app, stack_config):
        self.app = app
        self.config = stack_config
        self.name = stack_config["name"]
        self.namespace = app.namespace

    @property
    @functools.lru_cache(maxsize=32)
    def stack_name(self):
        """Cloudformation stack name."""
        if self.config.get("account_unique", False):
            parts = [self.app.name, self.name]
        else:
            parts = [self.app.namespace, self.name]
        return "-".join(parts)

    @property
    @functools.lru_cache(maxsize=32)
    def resource(self):
        resource_ = {}
        paginator = self.cfn.get_paginator("list_stack_resources")
        for page in paginator.paginate(StackName=self.stack_name):
            for output in page["StackResourceSummaries"]:
                resource_[output["LogicalResourceId"]] = output[
                    "PhysicalResourceId"
                ]
        return resource_

    ############################################################################
    # Public APIs
    #
    # signature of these methods are stable and will not break, it is encouraged
    # to use these APIs so as to get the best practice in your build pipelines.
    ############################################################################
    def stack_exists(self):
        """Check whether this stack exists."""
        try:
            self.cfn.describe_stacks(StackName=self.stack_name)
            return True
        except ClientError as error:
            code = error.response["Error"]["Code"]
            message = error.response["Error"]["Message"]
            if code == "ValidationError" and message.endswith("does not exist"):
                return False
            raise

    @property
    @functools.lru_cache(maxsize=32)
    def get_stack_resource(self, stack_name, logical_name):
        """Get the pri of a resource by its logical_name in a stack."""
        response = self.cfn.describe_stack_resource(
            StackName=stack_name, LogicalResourceId=logical_name
        )
        return response["StackResourceDetail"]["PhysicalResourceId"]

    def get_stack_ssm(self):
        """Get all parameters created in this stack."""
        parameters = self.ssm.get_parameters_by_path(
            Path=f"/apps/{self.stack_name}/", WithDecryption=True
        )["Parameters"]
        return {p["Name"].split("/")[-1]: p["Value"] for p in parameters}

    def create_changeset(self, template, params):
        """Create a changeset."""
        name = f"{self.stack_name}-{datetime.now().strftime('%F-%H-%M-%S')}"
        kwargs = {
            "StackName": self.stack_name,
            "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_AUTO_EXPAND"],
            "Parameters": [
                {"ParameterKey": param, "ParameterValue": params[param]}
                for param in params
            ],
            "Tags": self.app.tags,
            "ChangeSetName": name,
            "ChangeSetType": "UPDATE" if self.stack_exists() else "CREATE",
        }
        if os.path.isfile(template):
            with open(template) as fobj:
                kwargs["TemplateBody"] = fobj.read()
        elif template.startswith("http"):
            kwargs["TemplateURL"] = template
        else:
            raise RuntimeError(f"Bad template: {template}.")
        logging.info(f"Creating changeset {name}.")
        self.cfn.create_change_set(**kwargs)
        return name

    def describe_changeset(self, name):
        """Wait till a changeset is available and return it's changes."""
        kwargs = {"ChangeSetName": name, "StackName": self.stack_name}

        start = datetime.now()
        while True:
            # Change set should be ready within seconds.
            sleep(start, 60)
            response = self.cfn.describe_change_set(**kwargs)
            status = response["Status"]
            exec_status = response["ExecutionStatus"]

            if status == "FAILED":
                reason = response["StatusReason"]
                # It just happened that AWS can give two reasons for this.
                if "didn't contain changes" in reason:
                    raise EmptyChangeSetError
                if reason == "No updates are to be performed.":
                    raise EmptyChangeSetError
                raise RuntimeError(
                    f"Failed to create changeset for {self.stack_name}: "
                    f"{reason}"
                )

            if status == "CREATE_COMPLETE" and exec_status == "AVAILABLE":
                if not response.get("NextToken"):
                    return response["Changes"]
                changes = response["Changes"]
                kwargs["NextToken"] = response["NextToken"]
                break
            logging.info(f"Status of changeset is `{status}`.")

        while True:
            response = self.cfn.describe_change_set(**kwargs)
            changes += response["Changes"]
            if not response.get("NextToken"):
                break
            kwargs["NextToken"] = response["NextToken"]
        return changes

    def execute_changeset(self, name, timeout):
        """Execute a changeset."""
        logging.info(f"Executing changeset `{name}`.")
        self.cfn.execute_change_set(
            ChangeSetName=name, StackName=self.stack_name
        )

        start = datetime.now()
        while True:
            sleep(start, timeout)
            response = self.cfn.describe_stacks(StackName=self.stack_name)
            status = response["Stacks"][0]["StackStatus"]
            if status.endswith("FAILED") or status.endswith("COMPLETE"):
                break
            logging.info(f"Waiting till stack operation completes: {status}.")

        logging.info(f"Stack operation finished: {status}")
        bad_statuses = [
            "UPDATE_ROLLBACK_COMPLETE",
            "ROLLBACK_COMPLETE",
            "DELETE_COMPLETE",
        ]
        if status in bad_statuses:
            raise RuntimeError("Failed to create/update stack.")

    def deploy_stack(self, template, params=None, timeout=3600):
        """Deploy stack changes by creating a changeset."""
        logging.info(f"Building/Updating {self.name} stack.")
        if params is None:
            params = {}
        name = self.create_changeset(template, params)
        try:
            changes = self.describe_changeset(name)
        except EmptyChangeSetError:
            logging.info(f"No change in {self.stack_name} stack.")
            return
        logging.info(
            f"Changes in changeset `{name}`: \n{format_changes(changes)}"
        )
        self.execute_changeset(name, timeout)
        return changes

    def delete_stack(self):
        """Remove this stack."""
        if not self.stack_exists():
            # If stack does not exists, ignore it.
            logging.info(f"Stack {self.stack_name} does not exists.")
            return

        response = self.cfn.describe_stacks(StackName=self.stack_name)
        stack_id = response["Stacks"][0]["StackId"]
        logging.info(f"Removing stack: {stack_id}.")
        self.cfn.delete_stack(StackName=stack_id)

        start = datetime.now()
        while True:
            sleep(start)
            response = self.cfn.describe_stacks(StackName=stack_id)
            status = response["Stacks"][0]["StackStatus"]
            if status.endswith("FAILED") or status.endswith("COMPLETE"):
                break
            logging.info(f"Waiting till stack operation completes: {status}.")

        if status != "DELETE_COMPLETE":
            raise RuntimeError("Failed to delete stack.")
        logging.info(f"Stack Removed: {stack_id}.")
        return stack_id

    def rollback(self, build):
        """
        Rollback a stack to a previous version.

        As build is the only parameter we need for a deployment,
        rolling back is done by deploying the previous version.
        """
        self.deploy(build)

    ############################################################################
    # commands
    ############################################################################
    def delete(self):
        """Remove a stack."""
        self.delete_stack()

    def deploy(self, build):
        """
        Deploy a stack.

        The inheriting class must explicitly define this method.
        """
        raise NotImplementedError



#!/usr/bin/env python3
# encoding=utf8
"""Utility functions in Ennio."""
from datetime import datetime
from functools import wraps
import logging
import os
import sys
import time

import boto3
from botocore.exceptions import NoCredentialsError, ClientError


def setup_logging():
    """Logging setup"""
    logging_kwargs = {
        "stream": sys.stdout,
        "format": "[%(asctime)s][%(levelname)s] %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
    }
    if "ENNIO_DEBUG" not in os.environ:
        # To enable debugging output, set DEBUG to any value in env var.
        logging.getLogger("boto3").setLevel(logging.WARNING)
        logging.getLogger("botocore").setLevel(logging.WARNING)
        logging.basicConfig(level=logging.INFO, **logging_kwargs)
    else:
        logging.basicConfig(level=logging.DEBUG, **logging_kwargs)


def sleep(start, timeout=None):
    """sleep with increasing intervals."""
    since_start = (datetime.now() - start).seconds

    if timeout is not None:
        if since_start > timeout:
            raise RuntimeError(f"Operation timeout in {timeout} seconds.")

    interval = int((since_start ** 0.5) * 2.5) + 4
    logging.debug(f"Sleeping {interval} seconds.")
    time.sleep(interval)


def require_aws(func):
    """Decorator to verify AWS session."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not hasattr(boto3, "caller_identity"):
            try:
                boto3.setup_default_session(region_name="ap-southeast-2")
                client = boto3.client("sts")
                # We add an attibute `caller_identity` to the boto3 object,
                # this will later be used in `EnnioApplication.account_id`.
                boto3.caller_identity = client.get_caller_identity()
            except (NoCredentialsError, ClientError):
                print("Please login to AWS.", file=sys.stderr)
                sys.exit(1)
        return func(*args, **kwargs)

    return wrapper


def format_changes(changes):
    """Format changes so it will look better."""
    parts = []
    logging.debug(f"Detailed changes: {changes}")
    for change in changes:
        change_ = change["ResourceChange"]
        line = (
            f"[{change_['Action'].upper()}] "
            f"{change_['LogicalResourceId']}({change_['ResourceType']})"
        )
        if change_["Details"]:
            line += f":\n\t{change_['Details']}"
        parts.append(line)
    return "\n".join(parts)


class EmptyChangeSetError(BaseException):
    """Raised when no change needed during stack updates."""


class InvalidConfigError(BaseException):
    """Raised when we have an invalid config file."""


class LazyBoto3Client:
    """A lazy boto3 client so we will only create the client when we use it."""

    def __init__(self, name):
        self.name = name
        self.client = None

    @require_aws
    def __get__(self, obj, *args):
        if self.client is None:
            self.client = boto3.client(self.name)
        return self.client



from setuptools import find_packages, setup


VERSION = "0.1.2"


with open("README.md") as fobj:
    long_description = fobj.read()

if __name__ == "__main__":
    setup(
        name="ennio",
        version=VERSION,
        author="Sportsbet",
        author_email="engineering@sportsbet.com.au",
        license="Apache-2.0",
        url="https://github.com/sportsbet/ennio",
        description="interstack orchestration framework for AWS CloudFormation",
        long_description=long_description,
        long_description_content_type="text/markdown",
        packages=find_packages(),
        install_requires=["PyYAML", "boto3", "jinja2"],
        classifiers=[
            "Development Status :: 4 - Beta",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
        ],
    )
